// Lisp2D - Mutex auto init/destroy with statistics
// ©2018-23 by Alex Glebe

# include <stdio.h> // stderr
# include <wchar.h> // fputws
# include <iso646.h> // not_eq
# include <pthread.h>
# include <errno.h> // EBUSY
# include "mutex2/type.h"
# include "mutex2/struct.h"
# include "mutex2/public.h"
# include "main.h"
# include "excc.h"
  
void  Mutex2Lock  ( Mutex2 * const mp ) {
  if  ( mp ->  s not_eq Lisp2DObjectStateConstructed ) {
    Lisp2DExceptionStringGlobal . str = (wchar_t const(*)[])&
      L"Mutex2Lock:mutex is not live" ;
    longjmp ( * Lisp2DJBglobal  , Lisp2DExcTypeString ) ;
  }
  int e ;  
# ifdef LISP2D_BIMSTAT
  e = pthread_mutex_trylock ( & ( mp ->  m ) ) ;
  if  ( e ==  0 ) {
    ++  ( mp -> statfast  ) ;
    return  ;
  }
  if ( e == EBUSY ) {
    ++ ( mp -> statwait ) ;
    e = pthread_mutex_lock  ( & mp -> m ) ;
  }
# else
  e = pthread_mutex_lock  ( & mp -> m ) ;
# endif
  switch  ( e ) {
  case  0 :
    return  ;
  case  EAGAIN  :
    Lisp2DExceptionStringGlobal . str = (wchar_t const(*)[])&
      L"Mutex2Lock:The mutex could not be acquired because the"
      L" maximum number of recursive locks for mutex has been exceeded."  ;
    break ;
  case  EINVAL  :
    Lisp2DExceptionStringGlobal . str = (wchar_t const(*)[])&
      L"Mutex2Lock:The mutex was created with the protocol attribute having "
      L"the value PTHREAD_PRIO_PROTECT and the calling thread's priority is "
      L"higher than the mutex's current priority ceiling."  ;
    break ;
  case  ENOTRECOVERABLE :
    Lisp2DExceptionStringGlobal . str = (wchar_t const(*)[])&
      L"Mutex2Lock:The state protected by the mutex is not recoverable."  ;
    break ;
  case  EOWNERDEAD :
    Lisp2DExceptionStringGlobal . str = (wchar_t const(*)[])&
      L"Mutex2Lock:The mutex is a robust mutex and the process containing "
      L"the previous owning thread terminated while  holding  the mutex  lock."
      L" The  mutex  lock shall be acquired by the calling thread and it is "
      L"up to the new owner to make the state consistent." ;
    break ;
  case  EDEADLK :
    Lisp2DExceptionStringGlobal . str = (wchar_t const(*)[])&
      L"Mutex2Lock:The mutex type is PTHREAD_MUTEX_ERRORCHECK and the current "
      L"thread already owns the mutex." ;
    break ;  
  default :
    Lisp2DExceptionStringGlobal . str = (wchar_t const(*)[])&L"Mutex2Lock:?" ;
  }
  longjmp ( * Lisp2DJBglobal , Lisp2DExcTypeString ) ;
}

void  Mutex2Unlock  ( Mutex2 * const mp ) {
  if  ( mp ->  s not_eq Lisp2DObjectStateConstructed ) {
    Lisp2DExceptionStringGlobal . str = (wchar_t const(*)[])&
      L"Mutex2Unlock:mutex is not live" ;
    longjmp ( * Lisp2DJBglobal , Lisp2DExcTypeString ) ;
  }
  int const e = pthread_mutex_unlock  ( & mp -> m ) ;
  switch  ( e ) {
  case  0 :
    return  ;
  case  EPERM  :
    Lisp2DExceptionStringGlobal . str = (wchar_t const(*)[])&
      L"Mutex2Unlock:The mutex type is PTHREAD_MUTEX_ERRORCHECK or "
      L"PTHREAD_MUTEX_RECURSIVE, or the mutex is a robust mutex,  and "
      L"the current thread does not own the mutex." ;
    break ;
  default :
    Lisp2DExceptionStringGlobal . str = (wchar_t const(*)[])&L"Mutex2Unlock:?" ;
  }
  longjmp ( * Lisp2DJBglobal , Lisp2DExcTypeString ) ;
}

void  Mutex2Constructor  ( Mutex2 * const mp , pthread_mutexattr_t const * const attr ) {
  if  ( mp ->  s == Lisp2DObjectStateConstructed ) {
    Lisp2DExceptionStringGlobal . str = (wchar_t const(*)[])&
      L"Mutex2Init:double construction" ;
    longjmp ( * Lisp2DJBglobal  , Lisp2DExcTypeString ) ;
  }
  int const e = pthread_mutex_init  ( & mp -> m , attr ) ;
  switch  ( e ) {
  case  0 :
    mp -> s = Lisp2DObjectStateConstructed ;
# ifdef  LISP2D_BIMSTAT
    mp  ->  statfast = 0 ;
    mp  ->  statwait = 0 ;
# endif
    return  ;
  case  EAGAIN  :
    Lisp2DExceptionStringGlobal . str = (wchar_t const(*)[])&
      L"Mutex2Init:The system lacked the necessary resources (other than "
      L"memory) to initialize another mutex."  ;
    break ;
  case  ENOMEM  :
    Lisp2DExceptionStringGlobal . str = (wchar_t const(*)[])&
      L"Mutex2Init:Insufficient memory exists to initialize the mutex." ;
    break ;
  case  EPERM :
    Lisp2DExceptionStringGlobal . str = (wchar_t const(*)[])&
      L"Mutex2Init:The caller does not have the privilege to perform the "
      L"operation."  ;
    break ;
  case  EINVAL :
    Lisp2DExceptionStringGlobal . str = (wchar_t const(*)[])&
      L"Mutex2Init:The attributes object referenced by attr has the robust "
      L"mutex attribute set without the process-shared attribute being set." ;
    break ;
  case  EBUSY :
    Lisp2DExceptionStringGlobal . str = (wchar_t const(*)[])&
      L"Mutex2Init:detects that the value specified  by  the  mutex argument "
      L"to  pthread_mutex_init()  refers  to an already initialized mutex." ;
    break ;  
  default :
    Lisp2DExceptionStringGlobal . str = (wchar_t const(*)[])&L"Mutex2Init:?" ; 
  }
  longjmp ( * Lisp2DJBglobal  , Lisp2DExcTypeString ) ;
}

void  Mutex2Destructor  ( Mutex2 * const mp ) {
  if  ( mp ->  s not_eq Lisp2DObjectStateConstructed ) {
    Lisp2DExceptionStringGlobal . str = (wchar_t const(*)[])&
      L"Mutex2Destroy:is not live"  ;
    longjmp ( * Lisp2DJBglobal  , Lisp2DExcTypeString ) ;
  }
  int const e = pthread_mutex_destroy  ( & mp -> m ) ;
  switch  ( e ) {
  case 0 :
    mp -> s =  Lisp2DObjectStateDestructed ;
    return ;
  case EINVAL :
    Lisp2DExceptionStringGlobal . str = (wchar_t const(*)[])&
      L"Mutex2Destroy:detects that the value specified by the mutex argument "
      L"to pthread_mutex_destroy() does not refer to an initialized mutex."  ;
    break ;
  case EBUSY :
    Lisp2DExceptionStringGlobal . str = (wchar_t const(*)[])&
      L"Mutex2Destroy:locked  mutex  or  a  mutex that is referenced (for "
      L"example, while being used in a pthread_cond_timedwait() or "
      L"pthread_cond_wait()) by another thread." ;
    break ;  
  default :
    Lisp2DExceptionStringGlobal . str = (wchar_t const(*)[])&
      L"Mutex2Destroy:?"  ;
  }
  mp -> s =  Lisp2DObjectStateRaw ;
  longjmp ( * Lisp2DJBglobal  , Lisp2DExcTypeString ) ;
}

pthread_mutex_t * Mutex2Addr ( Mutex2 * const mep ) {
  if  ( mep ->  s not_eq  Lisp2DObjectStateConstructed  ) {
    Lisp2DExceptionStringGlobal . str = (wchar_t const(*)[])&
      L"Mutex2Addr:mutex is not live" ;
    longjmp ( * Lisp2DJBglobal , Lisp2DExcTypeString ) ;
  }
  return & ( mep -> m ) ;
}
  
bool  Mutex2IsBadTryLock  ( Mutex2 * const mp ) {
  if  ( mp ->  s not_eq Lisp2DObjectStateConstructed ) {
    Lisp2DExceptionStringGlobal . str = (wchar_t const(*)[])&
      L"Mutex2TryLock:mutex is not live" ;
    longjmp ( * Lisp2DJBglobal  , Lisp2DExcTypeString ) ;
  }
  int const e = pthread_mutex_trylock  ( & mp -> m ) ;
  switch  ( e ) {
  case  0 :
# ifdef LISP2D_BIMSTAT
    ++  ( mp -> statfast  ) ;
# endif    
    return  false ;
  case  EBUSY :
# ifdef LISP2D_BIMSTAT
    ++  ( mp -> statwait ) ;
# endif
    return  true  ;
  case  EAGAIN  :
    Lisp2DExceptionStringGlobal . str = (wchar_t const(*)[])&
      L"Mutex2TryLock:The mutex could not be acquired because the maximum "
      L"number of recursive locks for mutex has been exceeded." ;
    break ;
  case  EINVAL  :
    Lisp2DExceptionStringGlobal . str = (wchar_t const(*)[])&
      L"Mutex2TryLock:The mutex was created with the protocol attribute "
      L"having the value PTHREAD_PRIO_PROTECT and the calling thread's "
      L"priority is higher than the mutex's current priority ceiling." ;
    break ;
  case  ENOTRECOVERABLE :
    Lisp2DExceptionStringGlobal . str = (wchar_t const(*)[])&
      L"Mutex2TryLock:The state protected by the mutex is not recoverable." ;
    break ;
  case  EOWNERDEAD :
    Lisp2DExceptionStringGlobal . str = (wchar_t const(*)[])&
      L"Mutex2TryLock:The mutex is a robust mutex and the process containing "
      L"the previous owning thread terminated while  holding  the mutex  lock. "
      L"The mutex  lock shall be acquired by the calling thread and it is up "
      L"to the new owner to make the state consistent."  ;
    break ;
  default :
    Lisp2DExceptionStringGlobal . str = (wchar_t const(*)[])&
      L"Mutex2TryLock:?"  ;
  }
  longjmp ( * Lisp2DJBglobal , Lisp2DExcTypeString ) ;
}    
  
void  Mutex2SetStateRaw  ( Mutex2 * const mp ) {
  mp -> s =  Lisp2DObjectStateRaw ;
}

# ifdef LISP2D_BIMSTAT  
size_t  Mutex2StatWait  ( Mutex2 const * const mep ) {
  return  mep -> statwait ;
}
# endif  
